import asyncio
from asyncio import Semaphore
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Awaitable, Callable, List, Optional, Type, TypeVar
from typing_extensions import TypeAlias

from loguru import logger

TestFunc: TypeAlias = Callable[[], Awaitable[Any]]

T = TypeVar("T")
TF = TypeVar("TF", bound=TestFunc)

tests: List[TestFunc] = []
before_exit_hooks: List[TestFunc] = []


def mark_test(func: TF) -> TF:
    tests.append(func)
    return func


def mark_before_exit(func: TF) -> TF:
    before_exit_hooks.append(func)
    return func


def wrapper_log(func: TestFunc) -> TestFunc:
    async def wrapper():
        logger.opt(colors=True).info(f"Running test <y>{func.__name__}</y>")
        try:
            return await func()
        except Exception:
            logger.exception(f"Test {func.__name__} failed")
            raise

    return wrapper


async def run_tests():
    from cookit import with_semaphore

    try:
        sem = Semaphore(8)
        await asyncio.gather(
            *(
                with_semaphore(sem)(
                    wrapper_log(x),
                )()
                for x in tests
            ),
        )
    except Exception:  # noqa: S110
        pass
    else:
        logger.success("All tests passed")
    finally:
        end_sem = Semaphore(8)
        await asyncio.gather(
            *(
                with_semaphore(end_sem)(
                    logger.catch(
                        message=f"Before exit hook {x.__name__} raised exception",
                    )(x),
                )()
                for x in before_exit_hooks
            ),
        )


def auto_import_tests(path: str, package: Optional[str]):
    from cookit import auto_import

    auto_import(
        Path(path).parent,
        package,
        lambda x: x.is_dir() or x.name.startswith("test_"),
    )


@contextmanager
def expect_exc(*exc: Type[BaseException]):
    try:
        yield
    except exc:
        return
    raise AssertionError



from pathlib import Path

from ..utils import expect_exc, mark_test
from .utils import get_page

html_text = (Path(__file__).parent / "test.html").read_text("u8")


@mark_test
async def test_screenshot_selector():
    from cookit.pw import screenshot_selector

    async with await get_page() as page:
        await page.set_content(html_text)
        await screenshot_selector(page, "h1")
        with expect_exc(ValueError):
            await screenshot_selector(page, "h5")


@mark_test
async def test_screenshot_html():
    from cookit.pw import screenshot_html

    async with await get_page() as page:
        await screenshot_html(page, html_text)

    async with await get_page() as page:
        await screenshot_html(page, html_text, "h1")

    async with await get_page() as page:
        with expect_exc(ValueError):
            await screenshot_html(page, html_text, "h5")



from typing import Optional

from jinja2 import Template
from playwright.async_api import Page

from ..screenshot import screenshot_html


async def make_template_renderer(template: Template, **template_kwargs):
    html = await template.render_async(**template_kwargs)

    async def render(page: Page, selector: Optional[str] = None, **screenshot_kwargs):
        return await screenshot_html(page, html, selector, **screenshot_kwargs)

    return render



import time
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Iterator,
    List,
    Literal,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    TypeVar,
    overload,
)
from typing_extensions import override

if TYPE_CHECKING:
    from os import stat_result

TF = TypeVar("TF", str, bytes)


class FileCacheManager(MutableMapping[str, TF]):
    @overload
    def __init__(
        self: "FileCacheManager[str]",
        cache_dir: Path,
        /,
        text_mode: Literal[True],
        max_size: Optional[int] = None,
        ttl: Optional[int] = None,
        encoding: str = "utf-8",
    ) -> None: ...
    @overload
    def __init__(
        self: "FileCacheManager[bytes]",
        cache_dir: Path,
        /,
        text_mode: bool = False,
        max_size: Optional[int] = None,
        ttl: Optional[int] = None,
    ) -> None: ...
    def __init__(
        self,
        cache_dir: Path,
        /,
        text_mode: bool = False,
        max_size: Optional[int] = None,
        ttl: Optional[int] = None,
        encoding: str = "utf-8",
    ):
        self.cache_dir = cache_dir
        self.text_mode = text_mode
        self.max_size = max_size
        self.ttl = ttl
        self.encoding = encoding

    def purge(self):
        if (not self.cache_dir.exists()) or ((not self.max_size) and (not self.ttl)):
            return

        files: List[Tuple[Path, "stat_result"]] = [
            (x, x.stat()) for x in self.cache_dir.iterdir() if x.is_file()
        ]
        files.sort(key=lambda x: x[1].st_mtime)
        files_len = len(files)

        should_unlink: Set[Path] = set()
        if self.max_size and files_len > self.max_size:
            should_unlink.update(x[0] for x in files[: files_len - self.max_size])

        if self.ttl:
            out_time = time.time() - self.ttl
            index = next(
                (i for i, x in enumerate(files) if x[1].st_mtime >= out_time),
                files_len,
            )
            should_unlink.update(path for path, _ in files[:index])

        for path in should_unlink:
            path.unlink(missing_ok=True)

    @override
    def __getitem__(self, key: str) -> TF:
        self.purge()
        if (not (path := (self.cache_dir / key)).exists()) or (not path.is_file()):
            raise KeyError(key)
        return path.read_text(self.encoding) if self.text_mode else path.read_bytes()  # type: ignore

    @override
    def __setitem__(self, key: str, value: TF) -> None:
        if not self.cache_dir.exists():
            self.cache_dir.mkdir(parents=True)
        path = self.cache_dir / key
        (
            path.write_text(value, self.encoding)  # type: ignore
            if self.text_mode
            else path.write_bytes(value)  # type: ignore
        )
        self.purge()

    @override
    def __delitem__(self, key: str) -> None:
        (self.cache_dir / key).unlink(missing_ok=True)

    @override
    def __iter__(self) -> Iterator[str]:
        self.purge()
        return (x.name for x in self.cache_dir.iterdir() if x.is_file())

    @override
    def __len__(self) -> int:
        self.purge()
        return len([x for x in self.cache_dir.iterdir() if x.is_file()])

    @override
    def __contains__(self, key: Any) -> bool:
        if not isinstance(key, str):
            return False
        self.purge()
        return (self.cache_dir / key).exists()



import asyncio
import shutil
from pathlib import Path


async def test_file_cache_manager():
    from cookit import FileCacheManager

    path = (
        Path(__file__).parent.parent.parent.parent
        / ".pytest_cache"
        / "common_data_test_cache"
    )

    try:
        bytes_cache = FileCacheManager(path, max_size=2, ttl=1)

        bytes_cache["test"] = b"test"
        assert "test" in bytes_cache
        assert bytes_cache.get("test") == b"test"

        bytes_cache["test"] = b"test test"
        assert "test" in bytes_cache
        assert bytes_cache.get("test") == b"test test"

        bytes_cache["test2"] = b"test2"
        assert "test2" in bytes_cache
        assert bytes_cache.get("test2") == b"test2"

        bytes_cache["test3"] = b"test3"
        assert "test3" in bytes_cache
        assert bytes_cache.get("test3") == b"test3"
        assert "test2" in bytes_cache
        assert bytes_cache.get("test2") == b"test2"
        assert "test" not in bytes_cache
        assert bytes_cache.get("test") is None

        await asyncio.sleep(0.5)

        bytes_cache["test4"] = b"test4"
        assert "test4" in bytes_cache
        assert bytes_cache.get("test4") == b"test4"
        assert "test" not in bytes_cache
        assert bytes_cache.get("test") is None
        assert "test2" not in bytes_cache
        assert bytes_cache.get("test2") is None

        await asyncio.sleep(0.5)

        assert "test" not in bytes_cache
        assert bytes_cache.get("test") is None
        assert "test2" not in bytes_cache
        assert bytes_cache.get("test2") is None
        assert "test3" not in bytes_cache
        assert bytes_cache.get("test3") is None
        assert "test4" in bytes_cache
        assert bytes_cache.get("test4") == b"test4"

        await asyncio.sleep(0.5)

        assert "test" not in bytes_cache
        assert bytes_cache.get("test") is None
        assert "test2" not in bytes_cache
        assert bytes_cache.get("test2") is None
        assert "test3" not in bytes_cache
        assert bytes_cache.get("test3") is None
        assert "test4" not in bytes_cache
        assert bytes_cache.get("test4") is None

        text_cache = FileCacheManager(path, text_mode=True, max_size=2, ttl=1)

        text_cache["test_text"] = "test测试测试"
        assert "test_text" in text_cache
        assert text_cache.get("test_text") == "test测试测试"
        assert bytes_cache.get("test_text") == "test测试测试".encode("utf-8")

    finally:
        shutil.rmtree(path)
